import * as WebSocket from 'ws';
import ClientManager from '../clientManager';
import ChannelManager from '../channelManager';
import PublicChannel from '../channels/types/publicChannel';
import PrivateChannel from '../channels/types/privateChannel';
import CustomChannel from '../channels/types/customChannel';
import PublicClient from '../clients/types/publicClient';
import PrivateClient from '../clients/types/privateClient';
import CustomClient from '../clients/types/customClient';

var messageManager = require('../messageManager');
var logger = require('../logger');

class ClientBase {
  ws: WebSocket;
  data: any;
  id: number;
  channel: PublicChannel|PrivateChannel|CustomChannel|null;
  clientManager: ClientManager;
  channelManager: ChannelManager;
  roles: Array<string>;

  constructor(data: any, ws: WebSocket, channelManager: ChannelManager, clientManager: ClientManager) {
    this.ws = ws;
    this.data = data;
    this.id = data.user_id;
    this.channel = null;
    this.clientManager = clientManager;
    this.channelManager = channelManager;
    this.roles = ['receiver']
  }

  getData() {
    return this.data;
  }

  type() {
    return this.data.client_type;
  }

  clientType() {
    return this.data.client;
  }

  connectToChannel(channel: PublicChannel|PrivateChannel|CustomChannel) {
    this.channel = channel;

    let messageListener = (data: any) => {
      logger.accessLog.info(`starting message transaction on channel ${channel.id}: `, {data: data});
      data = messageManager.prepareMessage(data, channel, this);

      if (!data.error) {
        if (data.message_type == 'broadcast') {
          channel.broadcastMessage(this, data);
        } else if (data.message_type == 'direct') {
          let to = this.clientManager.getClient(data.message.to)
          to.directMessage(data)
        } else if (data.message_type == 'changeChannel') {
          this.ws.removeListener('message', messageListener);
          this.channelManager.changeChannel(this, data);
        }

        logger.accessLog.info(`message transaction complete on channel ${channel.id}: `, {message: data});
      } else {
        logger.errorLog.info(`Validation failed, please review schema: ${channel.id}`, {data: {message: data, error: data.error}});
      }
    }

    this.ws.on('message', messageListener);

    this.ws.on('close', (reasonCode: string, description: string) => {
      logger.accessLog.info(`closing connection for client ${this.id}`);

      if (this.channel) {
        this.channel.removeClient(this.id);
      }

      this.clientManager.removeClient(this.id);
      logger.accessLog.info(`closed connection for client ${this.id}`);
    });
  }

  directMessage(message: any) {
    this.ws.send(JSON.stringify(message));
    logger.accessLog.info(`sent direct message to ${this.id}`, { data: { message: message }});
  }

  replaceWebSocket(ws: WebSocket) {
    this.ws.close();
    this.ws = ws;
  }
};

export default ClientBase;
